import java.util.Timer
import java.util.TimerTask
import scala.util.control.NonFatal
import scala.util.Try
import scala.concurrent.Future
import scala.concurrent.ExecutionContext.Implicits.global

trait Observable[T] {
  def subscribe(observer: Observer[T]): Subscription
  
  def map[U](f: T => U): Observable[U] = Observable[U] { obs =>
    subscribe(Observer[T](
        t => Try(f(t)).map(obs.onNext).recover { 
          case NonFatal(e) => obs.onError(e) }.get,
        obs.onError, obs.onComplete()))
  }
  
  def foreach(f: T => Unit) = subscribe(Observer(f))
  
  def take(count: Int) = Observable[T] { obs =>
    var n = 1
    val observer= Observer[T](
      t => if (n < count) { n += 1; obs.onNext(t) } else if (n == count) {
        obs.onNext(t); obs.onComplete()
      },
      obs.onError,
      obs.onComplete()
    ) 
    subscribe(observer)
  }
  
  def repeat: Observable[T] = Observable[T] { obs =>
    val subscription = MultipleAssignmentSubscription()
    lazy val observer: Observer[T] = Observer[T](obs.onNext, obs.onError, 
        subscription.subscription = this.subscribe(observer))
    subscription.subscription = subscribe(observer)
    subscription
  }
  
  def onErrorResumeNext(other: => Observable[T]) = Observable[T] { obs =>
    val subscription = MultipleAssignmentSubscription()
    lazy val observer: Observer[T] = Observer[T](obs.onNext, _ => subscription.subscription = other.subscribe(obs), obs.onComplete())
    subscription.subscription = subscribe(observer)
    subscription
  }
}

object Observable {
  def apply[T](create: Observer[T] => Subscription) = new Observable[T] {
    def subscribe(obs: Observer[T]) = {
      var unsubscribed = false
      lazy val observer: Observer[T] = Observer[T](
          t => if (!unsubscribed) obs.onNext(t),
          e => if (!unsubscribed) {
            subscription.unsubscribe()
            obs.onError(e)
          }, if (!unsubscribed) {
            subscription.unsubscribe()
            obs.onComplete()
          })
      lazy val underlying = create(observer)
      lazy val subscription = Subscription {        
        unsubscribed = true
        underlying.unsubscribe()
      }
      underlying; subscription
    }
  }
  
  def from[T](seq: Iterable[T])(implicit scheduler: Scheduler) = Observable[T] { obs =>
    scheduler.schedule {
      seq.foreach(obs.onNext)
      obs.onComplete()
    }
  }
  
  def interval(delay: Long) = Observable[Long] { obs =>
    val timer = new Timer
    var counter = 0l
    lazy val task: TimerTask = new TimerTask {
      def run() {
        try { obs.onNext(counter) }
        catch {
          case NonFatal(e) => s.unsubscribe(); obs.onError(e)
        }
        counter += 1
      }
    }
    timer.schedule(task, delay, delay)
    lazy val s = Subscription { timer.cancel() }
    s
  }
}
